@@ -575,9 +575,9 @@ int dtls_connection_init(struct dtls_connection *d, struct packet_stream *ps, in
 
 #if defined(SSL_OP_NO_QUERY_MTU)
 	SSL_CTX_set_options(d->ssl_ctx, SSL_OP_NO_QUERY_MTU);
-	SSL_set_mtu(d->ssl, 1500);
+	SSL_set_mtu(d->ssl, rtpe_config.dtls_mtu);
 #if defined(DTLS_set_link_mtu) || defined(DTLS_CTRL_SET_LINK_MTU) || OPENSSL_VERSION_NUMBER >= 0x10100000L
-	DTLS_set_link_mtu(d->ssl, 1500);
+	DTLS_set_link_mtu(d->ssl, rtpe_config.dtls_mtu);
 #endif
 #endif
 
@@ -701,7 +701,7 @@ int dtls(struct stream_fd *sfd, const str *s, const endpoint_t *fsin) {
 		return 0;
 
 	if (s)
-		__DBG("dtls packet input: len %u %02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
+		__DBG("dtls packet input: len %zu %02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
 			s->len,
 			(unsigned char) s->s[0], (unsigned char) s->s[1], (unsigned char) s->s[2], (unsigned char) s->s[3],
 			(unsigned char) s->s[4], (unsigned char) s->s[5], (unsigned char) s->s[6], (unsigned char) s->s[7],
@@ -729,7 +729,7 @@ int dtls(struct stream_fd *sfd, const str *s, const endpoint_t *fsin) {
 		/* connected! */
 		mutex_lock(&ps->out_lock); // nested lock!
 		if (dtls_setup_crypto(ps, d))
-			/* XXX ?? */ ;
+			{} /* XXX ?? */
 		mutex_unlock(&ps->out_lock);
 
 		if (PS_ISSET(ps, RTP) && PS_ISSET(ps, RTCP) && ps->rtcp_sibling
@@ -740,7 +740,7 @@ int dtls(struct stream_fd *sfd, const str *s, const endpoint_t *fsin) {
 			mutex_lock(&ps->rtcp_sibling->in_lock);
 			mutex_lock(&ps->rtcp_sibling->out_lock);
 			if (dtls_setup_crypto(ps->rtcp_sibling, d))
-				/* XXX ?? */ ;
+				{} /* XXX ?? */
 			mutex_unlock(&ps->rtcp_sibling->out_lock);
 			mutex_unlock(&ps->rtcp_sibling->in_lock);
 		}
@@ -768,11 +768,16 @@ int dtls(struct stream_fd *sfd, const str *s, const endpoint_t *fsin) {
 			buf[8], buf[9], buf[10], buf[11],
 			buf[12], buf[13], buf[14], buf[15]);
 
-		if (!fsin)
+		if (!fsin) {
 			fsin = &ps->endpoint;
+			if (fsin->port == 9)
+				fsin = NULL;
+		}
 
-		ilogs(srtp, LOG_DEBUG, "Sending DTLS packet");
-		socket_sendto(&sfd->socket, buf, ret, fsin);
+		if (fsin) {
+			ilogs(srtp, LOG_DEBUG, "Sending DTLS packet");
+			socket_sendto(&sfd->socket, buf, ret, fsin);
+		}
 	}
 
 	return 0;
@@ -816,7 +821,7 @@ void dtls_shutdown(struct packet_stream *ps) {
 		ps->dtls_cert = NULL;
 	}
 
-	crypto_reset(&ps->crypto);
+	call_stream_crypto_reset(ps);
 }
 
 void dtls_connection_cleanup(struct dtls_connection *c) {
