@@ -16,6 +16,10 @@
 #include <ifaddrs.h>
 #include <net/if.h>
 #include <netdb.h>
+#include <stdbool.h>
+#ifdef HAVE_MQTT
+#include <mosquitto.h>
+#endif
 
 #include "poller.h"
 #include "control_tcp.h"
@@ -49,10 +53,13 @@
 #include "jitter_buffer.h"
 #include "websocket.h"
 #include "codec.h"
+#include "mqtt.h"
+#include "janus.h"
 
 
 
 struct poller *rtpe_poller;
+struct poller_map *rtpe_poller_map;
 struct rtpengine_config initial_rtpe_config;
 
 static struct control_tcp *rtpe_tcp;
@@ -78,8 +85,15 @@ struct rtpengine_config rtpe_config = {
 	.redis_connect_timeout = 1000,
 	.media_num_threads = -1,
 	.dtls_rsa_key_size = 2048,
+	.dtls_mtu = 1200, // chrome default mtu
 	.dtls_signature = 256,
 	.max_dtx = 30,
+	.dtx_shift = 5,
+	.dtx_buffer = 10,
+	.dtx_lag = 100,
+	.mqtt_port = 1883,
+	.mqtt_keepalive = 30,
+	.mqtt_publish_interval = 5000,
 	.common = {
 		.log_levels = {
 			[log_level_index_internals] = -1,
@@ -330,7 +344,7 @@ static int redis_ep_parse(endpoint_t *ep, int *db, char **auth, const char *auth
 	char *sl;
 	long l;
 
-	sl = strchr(str, '@');
+	sl = strrchr(str, '@');
 	if (sl) {
 		*sl = 0;
 		*auth = g_strdup(str);
@@ -358,6 +372,32 @@ static int redis_ep_parse(endpoint_t *ep, int *db, char **auth, const char *auth
 }
 
 
+static void parse_cn_payload(str *out, char **in, const char *def, const char *name) {
+	if (!in || !*in) {
+		if (def)
+			str_init_dup(out, def);
+		return;
+	}
+
+	int len = g_strv_length(in);
+	if (len < 1)
+		die("Invalid CN payload specified (--%s)", name);
+	out->s = malloc(len);
+	for (int i = 0; i < len; i++) {
+		char *endp;
+		long p = strtol(in[i], &endp, 0);
+		if (endp == in[i] || *endp != '\0')
+			die("Invalid CN payload specified (--%s)", name);
+		if (p < 0 || p > 254)
+			die("Invalid CN payload specified (--%s)", name);
+		if (i == 0 && p > 127)
+			die("Invalid CN payload specified (--%s)", name);
+		out->s[i] = p;
+	}
+	out->len = len;
+}
+
+
 
 static void options(int *argc, char ***argv) {
 	AUTO_CLEANUP_GVBUF(if_a);
@@ -390,7 +430,15 @@ static void options(int *argc, char ***argv) {
 	AUTO_CLEANUP_GBUF(dtls_sig);
 	double silence_detect = 0;
 	AUTO_CLEANUP_GVBUF(cn_payload);
+	AUTO_CLEANUP_GVBUF(dtx_cn_params);
 	int debug_srtp = 0;
+	AUTO_CLEANUP_GBUF(amr_dtx);
+#ifdef HAVE_MQTT
+	AUTO_CLEANUP_GBUF(mqtt_publish_scope);
+#endif
+	AUTO_CLEANUP_GBUF(mos);
+
+	rwlock_lock_w(&rtpe_config.config_lock);
 
 	GOptionEntry e[] = {
 		{ "table",	't', 0, G_OPTION_ARG_INT,	&rtpe_config.kernel_table,		"Kernel table to use",		"INT"		},
@@ -425,6 +473,7 @@ static void options(int *argc, char ***argv) {
 		{ "redis-connect-timeout", 0, 0, G_OPTION_ARG_INT, &rtpe_config.redis_connect_timeout, "Sets a timeout in milliseconds for redis connections", "INT" },
 		{ "redis-delete-async", 'y', 0, G_OPTION_ARG_INT, &rtpe_config.redis_delete_async, "Enable asynchronous redis delete", NULL },
 		{ "redis-delete-async-interval", 'y', 0, G_OPTION_ARG_INT, &rtpe_config.redis_delete_async_interval, "Set asynchronous redis delete interval (seconds)", NULL },
+		{ "active-switchover", 0,0,G_OPTION_ARG_NONE,	&rtpe_config.active_switchover, "Use call activity as indicator of active/standby state", NULL },
 		{ "b2b-url",	'b', 0, G_OPTION_ARG_STRING,	&rtpe_config.b2b_url,	"XMLRPC URL of B2B UA"	,	"STRING"	},
 		{ "log-facility-cdr",0,  0, G_OPTION_ARG_STRING, &log_facility_cdr_s, "Syslog facility to use for logging CDRs", "daemon|local0|...|local7"},
 		{ "log-facility-rtcp",0,  0, G_OPTION_ARG_STRING, &log_facility_rtcp_s, "Syslog facility to use for logging RTCP", "daemon|local0|...|local7"},
@@ -432,6 +481,7 @@ static void options(int *argc, char ***argv) {
 		{ "log-facility-dtmf",0,  0, G_OPTION_ARG_STRING, &log_facility_dtmf_s, "Syslog facility to use for logging DTMF", "daemon|local0|...|local7"},
 		{ "dtmf-log-dest", 0,0,	G_OPTION_ARG_STRING,	&dtmf_udp_ep,	"Destination address for DTMF logging via UDP",	"IP46|HOSTNAME:PORT"	},
 		{ "dtmf-log-ng-tcp", 0,0,	G_OPTION_ARG_NONE,	&rtpe_config.dtmf_via_ng,	"DTMF logging via TCP NG protocol",	NULL },
+		{ "dtmf-no-suppress", 0,0,G_OPTION_ARG_NONE,	&rtpe_config.dtmf_no_suppress,	"Disable audio suppression during DTMF events",	NULL },
 #endif
 		{ "log-format",	0, 0,	G_OPTION_ARG_STRING,	&log_format,	"Log prefix format",		"default|parsable"},
 		{ "xmlrpc-format",'x', 0, G_OPTION_ARG_INT,	&rtpe_config.fmt,	"XMLRPC timeout request format to use. 0: SEMS DI, 1: call-id only, 2: Kamailio",	"INT"	},
@@ -469,6 +519,7 @@ static void options(int *argc, char ***argv) {
 		{ "jb-clock-drift",0,0,	G_OPTION_ARG_NONE,	&rtpe_config.jb_clock_drift,"Compensate for source clock drift",NULL },
 		{ "debug-srtp",0,0,	G_OPTION_ARG_NONE,	&debug_srtp,		"Log raw encryption details for SRTP",	NULL },
 		{ "dtls-rsa-key-size",0, 0,	G_OPTION_ARG_INT,&rtpe_config.dtls_rsa_key_size,"Size of RSA key for DTLS",	"INT"		},
+		{ "dtls-mtu",0, 0,	G_OPTION_ARG_INT,&rtpe_config.dtls_mtu,"DTLS MTU",	"INT"		},
 		{ "dtls-ciphers",0,  0,	G_OPTION_ARG_STRING,	&rtpe_config.dtls_ciphers,"List of ciphers for DTLS",		"STRING"	},
 		{ "dtls-signature",0,  0,G_OPTION_ARG_STRING,	&dtls_sig,		"Signature algorithm for DTLS",		"SHA-256|SHA-1"	},
 		{ "listen-http", 0,0,	G_OPTION_ARG_STRING_ARRAY,&rtpe_config.http_ifs,"Interface for HTTP and WS",	"[IP46|HOSTNAME:]PORT"},
@@ -476,13 +527,41 @@ static void options(int *argc, char ***argv) {
 		{ "https-cert", 0,0,	G_OPTION_ARG_STRING,	&rtpe_config.https_cert,"Certificate for HTTPS and WSS","FILE"},
 		{ "https-key", 0,0,	G_OPTION_ARG_STRING,	&rtpe_config.https_key,	"Private key for HTTPS and WSS","FILE"},
 		{ "http-threads", 0,0,	G_OPTION_ARG_INT,	&rtpe_config.http_threads,"Number of worker threads for HTTP and WS","INT"},
+		{ "software-id", 0,0,	G_OPTION_ARG_STRING,	&rtpe_config.software_id,"Identification string of this software presented to external systems","STRING"},
+		{ "poller-per-thread", 0,0,	G_OPTION_ARG_NONE,	&rtpe_config.poller_per_thread,	"Use poller per thread",	NULL },
 #ifdef WITH_TRANSCODING
 		{ "dtx-delay",	0,0,	G_OPTION_ARG_INT,	&rtpe_config.dtx_delay,	"Delay in milliseconds to trigger DTX handling","INT"},
 		{ "max-dtx",	0,0,	G_OPTION_ARG_INT,	&rtpe_config.max_dtx,	"Maximum duration of DTX handling",	"INT"},
+		{ "dtx-buffer",	0,0,	G_OPTION_ARG_INT,	&rtpe_config.dtx_buffer,"Maxmium number of packets held in DTX buffer",	"INT"},
+		{ "dtx-lag",	0,0,	G_OPTION_ARG_INT,	&rtpe_config.dtx_lag,	"Maxmium time span in milliseconds held in DTX buffer",	"INT"},
+		{ "dtx-shift",	0,0,	G_OPTION_ARG_INT,	&rtpe_config.dtx_shift,	"Length of time (in ms) to shift DTX buffer after over/underflow",	"INT"},
+		{ "dtx-cn-params",0,0,	G_OPTION_ARG_STRING_ARRAY,&dtx_cn_params,	"Parameters for CN generated from DTX","INT INT INT ..."},
+		{ "amr-dtx", 0,0,	G_OPTION_ARG_STRING,	&amr_dtx,		"DTX mechanism to use for AMR and AMR-WB","native|CN"},
 		{ "silence-detect",0,0,	G_OPTION_ARG_DOUBLE,	&silence_detect,	"Audio level threshold in percent for silence detection","FLOAT"},
 		{ "cn-payload",0,0,	G_OPTION_ARG_STRING_ARRAY,&cn_payload,		"Comfort noise parameters to replace silence with","INT INT INT ..."},
 		{ "reorder-codecs",0,0,	G_OPTION_ARG_NONE,	&rtpe_config.reorder_codecs,"Reorder answer codecs based on sender preference",NULL},
 #endif
+#ifdef HAVE_MQTT
+		{ "mqtt-host",0,0,	G_OPTION_ARG_STRING,	&rtpe_config.mqtt_host,	"Mosquitto broker host or address",	"HOST|IP"},
+		{ "mqtt-port",0,0,	G_OPTION_ARG_INT,	&rtpe_config.mqtt_port,	"Mosquitto broker port number",		"INT"},
+		{ "mqtt-id",0,0,	G_OPTION_ARG_STRING,	&rtpe_config.mqtt_id,	"Mosquitto client ID",			"STRING"},
+		{ "mqtt-keepalive",0,0,	G_OPTION_ARG_INT,	&rtpe_config.mqtt_keepalive,"Seconds between mosquitto keepalives","INT"},
+		{ "mqtt-user",0,0,	G_OPTION_ARG_STRING,	&rtpe_config.mqtt_user,	"Username for mosquitto auth",		"USERNAME"},
+		{ "mqtt-pass",0,0,	G_OPTION_ARG_STRING,	&rtpe_config.mqtt_pass,	"Password for mosquitto auth",		"PASSWORD"},
+		{ "mqtt-cafile",0,0,	G_OPTION_ARG_STRING,	&rtpe_config.mqtt_cafile,"CA file for mosquitto auth",		"FILE"},
+		{ "mqtt-capath",0,0,	G_OPTION_ARG_STRING,	&rtpe_config.mqtt_capath,"CA path for mosquitto auth",		"PATH"},
+		{ "mqtt-certfile",0,0,	G_OPTION_ARG_STRING,	&rtpe_config.mqtt_certfile,"Certificate file for mosquitto auth","FILE"},
+		{ "mqtt-keyfile",0,0,	G_OPTION_ARG_STRING,	&rtpe_config.mqtt_keyfile,"Key file for mosquitto auth",	"FILE"},
+		{ "mqtt-publish-qos",0,0,G_OPTION_ARG_INT,	&rtpe_config.mqtt_publish_qos,"Mosquitto publish QoS",		"0|1|2"},
+		{ "mqtt-publish-topic",0,0,G_OPTION_ARG_STRING,	&rtpe_config.mqtt_publish_topic,"Mosquitto publish topic",	"STRING"},
+		{ "mqtt-publish-interval",0,0,G_OPTION_ARG_INT,	&rtpe_config.mqtt_publish_interval,"Publish timer interval",	"MILLISECONDS"},
+		{ "mqtt-publish-scope",0,0,G_OPTION_ARG_STRING,	&mqtt_publish_scope,	"Scope for published mosquitto messages","global|call|media"},
+#endif
+		{ "mos",0,0,		G_OPTION_ARG_STRING,	&mos,		"Type of MOS calculation","CQ|LQ"},
+#ifdef SO_INCOMING_CPU
+		{ "socket-cpu-affinity",0,0,G_OPTION_ARG_INT,	&rtpe_config.cpu_affinity,"CPU affinity for media sockets","INT"},
+#endif
+		{ "janus-secret", 0,0,	G_OPTION_ARG_STRING,	&rtpe_config.janus_secret,"Admin secret for Janus protocol","STRING"},
 
 		{ NULL, }
 	};
@@ -507,8 +586,11 @@ static void options(int *argc, char ***argv) {
 
 	if (!if_a)
 		die("Missing option --interface");
-	if (!listenps && !listenudps && !listenngs && !listenngtcps)
-		die("Missing option --listen-tcp, --listen-udp or --listen-ng or --listen-tcp-ng");
+	if (!listenps && !listenudps && !listenngs && !listenngtcps
+			&& !(rtpe_config.http_ifs && rtpe_config.http_ifs[0])
+			&& !(rtpe_config.https_ifs && rtpe_config.https_ifs[0]))
+		die("Missing option --listen-tcp, --listen-udp, --listen-ng, --listen-tcp-ng, "
+				"--listen-http, or --listen-https");
 
 	struct ifaddrs *ifas;
 	if (getifaddrs(&ifas)) {
@@ -534,9 +616,9 @@ static void options(int *argc, char ***argv) {
 
 			if ((errno == ERANGE && (uint_keyspace_db == ULONG_MAX)) ||
 			    (errno != 0 && uint_keyspace_db == 0)) {
-				ilog(LOG_ERR, "Fail adding keyspace '%.*s' to redis notifications; errono=%d\n", str_keyspace_db.len, str_keyspace_db.s, errno);
+				ilog(LOG_ERR, "Fail adding keyspace '" STR_FORMAT "' to redis notifications; errono=%d\n", STR_FMT(&str_keyspace_db), errno);
 			} else if (endptr == str_keyspace_db.s) {
-				ilog(LOG_ERR, "Fail adding keyspace '%.*s' to redis notifications; no digits found\n", str_keyspace_db.len, str_keyspace_db.s);
+				ilog(LOG_ERR, "Fail adding keyspace '" STR_FORMAT "' to redis notifications; no digits found\n", STR_FMT(&str_keyspace_db));
 			} else {
 				g_queue_push_tail(&rtpe_config.redis_subscribed_keyspaces, GUINT_TO_POINTER(uint_keyspace_db));
 			}
@@ -706,6 +788,11 @@ static void options(int *argc, char ***argv) {
 	if (rtpe_config.dtls_rsa_key_size < 0)
 		die("Invalid --dtls-rsa-key-size (%i)", rtpe_config.dtls_rsa_key_size);
 
+	if (rtpe_config.dtls_mtu < 576)
+		/* The Internet Protocol requires that hosts must be able to process IP datagrams of at least 576 bytes (for IPv4) or 1280 bytes (for IPv6).
+		However, this does not preclude link layers with an MTU smaller than this minimum MTU from conveying IP data. Internet IPv4 path MTU is 68 bytes.*/
+		die("Invalid --dtls-mtu (%i)", rtpe_config.dtls_mtu);
+
 	if (rtpe_config.jb_length < 0)
 		die("Invalid negative jitter buffer size");
 
@@ -714,26 +801,43 @@ static void options(int *argc, char ***argv) {
 		rtpe_config.silence_detect_int = (int) ((silence_detect / 100.0) * UINT32_MAX);
 	}
 
-	if (!cn_payload)
-		str_init_dup(&rtpe_config.cn_payload, "\x20");
-	else {
-		int len = g_strv_length(cn_payload);
-		if (len < 1)
-			die("Invalid CN payload specified");
-		rtpe_config.cn_payload.s = malloc(len);
-		for (int i = 0; i < len; i++) {
-			char *endp;
-			long p = strtol(cn_payload[i], &endp, 0);
-			if (endp == cn_payload[i] || *endp != '\0')
-				die("Invalid CN payload specified");
-			if (p < 0 || p > 254)
-				die("Invalid CN payload specified");
-			if (i == 0 && p > 127)
-				die("Invalid CN payload specified");
-			rtpe_config.cn_payload.s[i] = p;
-		}
-		rtpe_config.cn_payload.len = len;
+	parse_cn_payload(&rtpe_config.cn_payload, cn_payload, "\x20", "cn-payload");
+	parse_cn_payload(&rtpe_config.dtx_cn_params, dtx_cn_params, NULL, "dtx-cn-params");
+
+	if (amr_dtx) {
+		if (!strcasecmp(amr_dtx, "native")) {}
+		else if (!strcasecmp(amr_dtx, "CN"))
+			rtpe_config.amr_cn_dtx = 1;
+		else
+			die("Invalid --amr-dtx ('%s')", amr_dtx);
 	}
+
+	if (!rtpe_config.software_id)
+		rtpe_config.software_id = g_strdup_printf("rtpengine-%s", RTPENGINE_VERSION);
+	g_strcanon(rtpe_config.software_id, "QWERTYUIOPASDFGHJKLZXCVBNMqwertyuiopasdfghjklzxcvbnm1234567890-", '-');
+
+#ifdef HAVE_MQTT
+	if (mqtt_publish_scope) {
+		if (!strcmp(mqtt_publish_scope, "global"))
+			rtpe_config.mqtt_publish_scope = MPS_GLOBAL;
+		else if (!strcmp(mqtt_publish_scope, "call"))
+			rtpe_config.mqtt_publish_scope = MPS_CALL;
+		else if (!strcmp(mqtt_publish_scope, "media"))
+			rtpe_config.mqtt_publish_scope = MPS_MEDIA;
+		else
+			die("Invalid --mqtt-publish-scope option ('%s')", mqtt_publish_scope);
+	}
+#endif
+	if (mos) {
+		if (!strcasecmp(mos, "cq"))
+			rtpe_config.mos = MOS_CQ;
+		else if (!strcasecmp(mos, "lq"))
+			rtpe_config.mos = MOS_LQ;
+		else
+			die("Invalid --mos option ('%s')", mos);
+	}
+
+	rwlock_unlock_w(&rtpe_config.config_lock);
 }
 
 void fill_initial_rtpe_cfg(struct rtpengine_config* ini_rtpe_cfg) {
@@ -741,6 +845,8 @@ void fill_initial_rtpe_cfg(struct rtpengine_config* ini_rtpe_cfg) {
 	GList* l;
 	struct intf_config* gptr_data;
 
+	rwlock_lock_w(&rtpe_config.config_lock);
+
 	for(l = rtpe_config.interfaces.head; l ; l=l->next) {
 		gptr_data = g_slice_alloc0(sizeof(*gptr_data));
 		memcpy(gptr_data, (struct intf_config*)(l->data), sizeof(*gptr_data));
@@ -811,6 +917,8 @@ void fill_initial_rtpe_cfg(struct rtpengine_config* ini_rtpe_cfg) {
 
 	ini_rtpe_cfg->jb_length = rtpe_config.jb_length;
 	ini_rtpe_cfg->jb_clock_drift = rtpe_config.jb_clock_drift;
+
+	rwlock_unlock_w(&rtpe_config.config_lock);
 }
 
 static void
@@ -859,6 +967,18 @@ static void options_free(void) {
 	g_strfreev(rtpe_config.https_ifs);
 	g_free(rtpe_config.https_cert);
 	g_free(rtpe_config.https_key);
+	g_free(rtpe_config.software_id);
+	if (rtpe_config.cn_payload.s)
+		g_free(rtpe_config.cn_payload.s);
+	if (rtpe_config.dtx_cn_params.s)
+		g_free(rtpe_config.dtx_cn_params.s);
+	g_free(rtpe_config.mqtt_user);
+	g_free(rtpe_config.mqtt_pass);
+	g_free(rtpe_config.mqtt_cafile);
+	g_free(rtpe_config.mqtt_certfile);
+	g_free(rtpe_config.mqtt_keyfile);
+	g_free(rtpe_config.mqtt_publish_topic);
+	g_free(rtpe_config.janus_secret);
 
 	// free common config options
 	config_load_free(&rtpe_config.common);
@@ -882,6 +1002,11 @@ static void init_everything(void) {
 	g_type_init();
 #endif
 
+#ifdef HAVE_MQTT
+	if (mosquitto_lib_init() != MOSQ_ERR_SUCCESS)
+		die("failed to init libmosquitto");
+#endif
+
 	signals();
 	resources();
 	sdp_init();
@@ -899,7 +1024,10 @@ static void init_everything(void) {
 	dtmf_init();
 	jitter_buffer_init();
 	t38_init();
+	if (rtpe_config.mqtt_host && mqtt_init())
+		abort();
 	codecs_init();
+	janus_init();
 }
 
 
@@ -923,6 +1051,10 @@ no_kernel:
 	if (!rtpe_poller)
 		die("poller creation failed");
 
+	rtpe_poller_map = poller_map_new();
+	if (!rtpe_poller_map)
+		die("poller map creation failed");
+
 	dtls_timer(rtpe_poller);
 
 	if (call_init())
@@ -982,20 +1114,20 @@ no_kernel:
 				endpoint_print_buf(&rtpe_config.redis_write_ep));
 	}
 
-		if (!is_addr_unspecified(&rtpe_config.redis_ep.address)) {
-			rtpe_redis = redis_new(&rtpe_config.redis_ep, rtpe_config.redis_db, rtpe_config.redis_auth, rtpe_redis_write ? ANY_REDIS_ROLE : MASTER_REDIS_ROLE, rtpe_config.no_redis_required);
-			if (!rtpe_redis)
+	if (!is_addr_unspecified(&rtpe_config.redis_ep.address)) {
+		rtpe_redis = redis_new(&rtpe_config.redis_ep, rtpe_config.redis_db, rtpe_config.redis_auth, rtpe_redis_write ? ANY_REDIS_ROLE : MASTER_REDIS_ROLE, rtpe_config.no_redis_required);
+		if (!rtpe_redis)
 			die("Cannot start up without running Redis %s database! "
 					"See also NO_REDIS_REQUIRED parameter.",
 				endpoint_print_buf(&rtpe_config.redis_ep));
 
-			if (rtpe_config.redis_subscribed_keyspaces.length) {
-				rtpe_redis_notify = redis_new(&rtpe_config.redis_ep, rtpe_config.redis_db, rtpe_config.redis_auth, rtpe_redis_write ? ANY_REDIS_ROLE : MASTER_REDIS_ROLE, rtpe_config.no_redis_required);
-				if (!rtpe_redis_notify)
-					die("Cannot start up without running notification Redis %s database! "
-							"See also NO_REDIS_REQUIRED parameter.",
-						endpoint_print_buf(&rtpe_config.redis_ep));
-			}
+		if (rtpe_config.redis_subscribed_keyspaces.length) {
+			rtpe_redis_notify = redis_new(&rtpe_config.redis_ep, rtpe_config.redis_db, rtpe_config.redis_auth, rtpe_redis_write ? ANY_REDIS_ROLE : MASTER_REDIS_ROLE, rtpe_config.no_redis_required);
+			if (!rtpe_redis_notify)
+				die("Cannot start up without running notification Redis %s database! "
+						"See also NO_REDIS_REQUIRED parameter.",
+					endpoint_print_buf(&rtpe_config.redis_ep));
+		}
 
 		if (!rtpe_redis_write)
 			rtpe_redis_write = rtpe_redis;
@@ -1004,6 +1136,12 @@ no_kernel:
 	if (rtpe_config.num_threads < 1)
 		rtpe_config.num_threads = num_cpu_cores(4);
 
+	if (rtpe_config.cpu_affinity < 0) {
+		rtpe_config.cpu_affinity = num_cpu_cores(0);
+		if (rtpe_config.cpu_affinity <= 0)
+			die("Number of CPU cores is unknown, cannot auto-set socket CPU affinity");
+	}
+
 	if (websocket_init())
 		die("Failed to init websocket listener");
 
@@ -1019,8 +1157,22 @@ no_kernel:
 		gettimeofday(&redis_start, NULL);
 
 		// restore
-		if (redis_restore(rtpe_redis))
-			die("Refusing to continue without working Redis database");
+		if (rtpe_redis_notify) {
+			// active-active mode: the main DB has our own calls, while
+			// the "notify" DB has the "foreign" calls. "foreign" DB goes
+			// first as the "owned" DB can do a stray update back to Redis
+			for (GList *l = rtpe_config.redis_subscribed_keyspaces.head; l; l = l->next) {
+				int db = GPOINTER_TO_INT(l->data);
+				if (redis_restore(rtpe_redis_notify, true, db))
+					ilog(LOG_WARN, "Unable to restore calls from the active-active peer");
+			}
+			if (redis_restore(rtpe_redis_write, false, -1))
+				die("Refusing to continue without working Redis database");
+		}
+		else {
+			if (redis_restore(rtpe_redis, false, -1))
+				die("Refusing to continue without working Redis database");
+		}
 
 		// stop redis restore timer
 		gettimeofday(&redis_stop, NULL);
@@ -1048,45 +1200,61 @@ int main(int argc, char **argv) {
 
 	ilog(LOG_INFO, "Startup complete, version %s", RTPENGINE_VERSION);
 
-	thread_create_detach(sighandler, NULL);
-	thread_create_detach_prio(poller_timer_loop, rtpe_poller, rtpe_config.idle_scheduling, rtpe_config.idle_priority);
-	thread_create_detach_prio(load_thread, NULL, rtpe_config.idle_scheduling, rtpe_config.idle_priority);
+	thread_create_detach(sighandler, NULL, "signal handler");
+	thread_create_detach_prio(poller_timer_loop, rtpe_poller, rtpe_config.idle_scheduling,
+			rtpe_config.idle_priority, "poller timer");
+	thread_create_detach_prio(load_thread, NULL, rtpe_config.idle_scheduling, rtpe_config.idle_priority, "load monitor");
 
 	if (!is_addr_unspecified(&rtpe_config.redis_ep.address) && initial_rtpe_config.redis_delete_async)
-		thread_create_detach(redis_delete_async_loop, NULL);
+		thread_create_detach(redis_delete_async_loop, NULL, "redis async");
 
 	if (!is_addr_unspecified(&rtpe_config.redis_ep.address) && rtpe_redis_notify)
-		thread_create_detach(redis_notify_loop, NULL);
+		thread_create_detach(redis_notify_loop, NULL, "redis notify");
 
 	if (!is_addr_unspecified(&rtpe_config.graphite_ep.address))
-		thread_create_detach(graphite_loop, NULL);
+		thread_create_detach(graphite_loop, NULL, "graphite");
 
-	thread_create_detach(ice_thread_run, NULL);
+#ifdef HAVE_MQTT
+	if (mqtt_publish_scope() != MPS_NONE)
+		thread_create_detach(mqtt_loop, NULL, "mqtt");
+#endif
+
+	thread_create_detach(ice_thread_run, NULL, "ICE");
 
 	websocket_start();
 
 	service_notify("READY=1\n");
 
-	for (idx = 0; idx < rtpe_config.num_threads; ++idx)
-		thread_create_detach_prio(poller_loop, rtpe_poller, rtpe_config.scheduling, rtpe_config.priority);
+	for (idx = 0; idx < rtpe_config.num_threads; ++idx) {
+		if (!rtpe_config.poller_per_thread)
+			thread_create_detach_prio(poller_loop2, rtpe_poller, rtpe_config.scheduling, rtpe_config.priority, "poller");
+		else
+			thread_create_detach_prio(poller_loop, rtpe_poller_map, rtpe_config.scheduling, rtpe_config.priority, "poller");
+	}
+
+	if (!rtpe_config.poller_per_thread)
+		thread_create_detach_prio(poller_loop2, rtpe_poller, rtpe_config.scheduling, rtpe_config.priority, "poller");
 
 	if (rtpe_config.media_num_threads < 0)
 		rtpe_config.media_num_threads = rtpe_config.num_threads;
 	for (idx = 0; idx < rtpe_config.media_num_threads; ++idx) {
 #ifdef WITH_TRANSCODING
-		thread_create_detach_prio(media_player_loop, NULL, rtpe_config.scheduling, rtpe_config.priority);
+		thread_create_detach_prio(media_player_loop, NULL, rtpe_config.scheduling,
+				rtpe_config.priority, "media player");
 #endif
-		thread_create_detach_prio(send_timer_loop, NULL, rtpe_config.scheduling, rtpe_config.priority);
+		thread_create_detach_prio(send_timer_loop, NULL, rtpe_config.scheduling,
+				rtpe_config.priority, "send timer");
 		if (rtpe_config.jb_length > 0)
 			thread_create_detach_prio(jitter_buffer_loop, NULL, rtpe_config.scheduling,
-					rtpe_config.priority);
-		thread_create_detach_prio(codec_timers_loop, NULL, rtpe_config.scheduling, rtpe_config.priority);
+					rtpe_config.priority, "jitter buffer");
+		thread_create_detach_prio(codec_timers_loop, NULL, rtpe_config.scheduling,
+				rtpe_config.priority, "codec timer");
 	}
 
 
 	while (!rtpe_shutdown) {
 		usleep(100000);
-		threads_join_all(0);
+		threads_join_all(false);
 	}
 
         // free libevent
@@ -1102,7 +1270,7 @@ int main(int argc, char **argv) {
 	if (!is_addr_unspecified(&rtpe_config.redis_ep.address) && rtpe_redis_notify)
 		redis_async_event_base_action(rtpe_redis_notify, EVENT_BASE_LOOPBREAK);
 
-	threads_join_all(1);
+	threads_join_all(true);
 
 	if (!is_addr_unspecified(&rtpe_config.redis_ep.address) && initial_rtpe_config.redis_delete_async)
 		redis_async_event_base_action(rtpe_redis_write, EVENT_BASE_FREE);
@@ -1121,13 +1289,12 @@ int main(int argc, char **argv) {
 	jitter_buffer_init_free();
 	media_player_free();
 	codeclib_free();
-	statistics_free();
 	call_interfaces_free();
-	interfaces_free();
 	ice_free();
 	dtls_cert_free();
 	control_ng_cleanup();
 	codecs_cleanup();
+	statistics_free();
 
 	redis_close(rtpe_redis);
 	if (rtpe_redis_write != rtpe_redis)
@@ -1135,16 +1302,17 @@ int main(int argc, char **argv) {
 	redis_close(rtpe_redis_notify);
 
 	free_prefix();
-
 	options_free();
-
 	log_free();
+	janus_free();
 
 	obj_release(rtpe_cli);
 	obj_release(rtpe_udp);
 	obj_release(rtpe_tcp);
 	obj_release(rtpe_control_ng);
 	poller_free(&rtpe_poller);
+	poller_map_free(&rtpe_poller_map);
+	interfaces_free();
 
 	return 0;
 }
