@@ -25,7 +25,7 @@
 
 #ifdef WITH_TRANSCODING
 static struct timerthread media_player_thread;
-static MYSQL __thread *mysql_conn;
+static __thread MYSQL *mysql_conn;
 
 static void media_player_read_packet(struct media_player *mp);
 #endif
@@ -43,6 +43,8 @@ static void send_timer_send_lock(struct send_timer *st, struct codec_packet *cp)
 #ifdef WITH_TRANSCODING
 // called with call->master lock in W
 static unsigned int send_timer_flush(struct send_timer *st, void *ptr) {
+	if (!st)
+		return 0;
 	return timerthread_queue_flush(&st->ttq, ptr);
 }
 
@@ -96,6 +98,7 @@ static void __media_player_free(void *p) {
 	ssrc_ctx_put(&mp->ssrc_out);
 	mutex_destroy(&mp->lock);
 	obj_put(mp->call);
+	av_packet_free(&mp->pkt);
 }
 #endif
 
@@ -108,7 +111,7 @@ struct media_player *media_player_new(struct call_monologue *ml) {
 	uint32_t ssrc = 0;
 	while (ssrc == 0)
 		ssrc = ssl_random();
-	struct ssrc_ctx *ssrc_ctx = get_ssrc_ctx(ssrc, ml->call->ssrc_hash, SSRC_DIR_OUTPUT, ml);
+	struct ssrc_ctx *ssrc_ctx = get_ssrc_ctx(ssrc, ml->ssrc_hash, SSRC_DIR_OUTPUT, ml);
 	ssrc_ctx->next_rtcp = rtpe_now;
 
 	struct media_player *mp = obj_alloc0("media_player", sizeof(*mp), __media_player_free);
@@ -122,9 +125,9 @@ struct media_player *media_player_new(struct call_monologue *ml) {
 	mp->seq = ssl_random();
 	mp->ssrc_out = ssrc_ctx;
 
-	av_init_packet(&mp->pkt);
-	mp->pkt.data = NULL;
-	mp->pkt.size = 0;
+	mp->pkt = av_packet_alloc();
+	mp->pkt->data = NULL;
+	mp->pkt->size = 0;
 
 	return mp;
 #else
@@ -182,13 +185,17 @@ static void __send_timer_send_common(struct send_timer *st, struct codec_packet
 	if (!st->sink->selected_sfd)
 		goto out;
 
+	log_info_stream_fd(st->sink->selected_sfd);
+
 	struct rtp_header *rh = cp->rtp;
-	if (rh)
+	if (rh) {
 		ilog(LOG_DEBUG, "Forward to sink endpoint: %s%s:%d%s (RTP seq %u TS %u)",
 				FMT_M(sockaddr_print_buf(&st->sink->endpoint.address),
 				st->sink->endpoint.port),
 				ntohs(rh->seq_num),
 				ntohl(rh->timestamp));
+		codec_calc_jitter(cp->ssrc_out, ntohl(rh->timestamp), cp->clockrate, &rtpe_now);
+	}
 	else
 		ilog(LOG_DEBUG, "Forward to sink endpoint: %s%s:%d%s",
 				FMT_M(sockaddr_print_buf(&st->sink->endpoint.address),
@@ -259,7 +266,7 @@ void send_timer_push(struct send_timer *st, struct codec_packet *cp) {
 int media_player_setup(struct media_player *mp, const struct rtp_payload_type *src_pt) {
 	// find suitable output payload type
 	struct rtp_payload_type *dst_pt;
-	for (GList *l = mp->media->codecs_prefs_send.head; l; l = l->next) {
+	for (GList *l = mp->media->codecs.codec_prefs.head; l; l = l->next) {
 		dst_pt = l->data;
 		ensure_codec_def(dst_pt, mp->media);
 		if (dst_pt->codec_def && !dst_pt->codec_def->supplemental)
@@ -319,13 +326,14 @@ static int __ensure_codec_handler(struct media_player *mp, AVStream *avs) {
 	src_pt.encoding = src_pt.codec_def->rtpname_str;
 	src_pt.channels = avs->CODECPAR->channels;
 	src_pt.clock_rate = avs->CODECPAR->sample_rate;
-	codec_init_payload_type(&src_pt, mp->media);
+	codec_init_payload_type(&src_pt, MT_AUDIO);
 
 	if (media_player_setup(mp, &src_pt))
 		return -1;
 
 	mp->duration = avs->duration * 1000 * avs->time_base.num / avs->time_base.den;
 
+	payload_type_clear(&src_pt);
 	return 0;
 }
 
@@ -384,7 +392,7 @@ static void media_player_read_packet(struct media_player *mp) {
 	if (!mp->fmtctx)
 		return;
 
-	int ret = av_read_frame(mp->fmtctx, &mp->pkt);
+	int ret = av_read_frame(mp->fmtctx, mp->pkt);
 	if (ret < 0) {
 		if (ret == AVERROR_EOF) {
 			if (mp->repeat > 1){
@@ -396,7 +404,7 @@ static void media_player_read_packet(struct media_player *mp) {
 				ret = av_seek_frame(mp->fmtctx, -1, 0, 0);
 				if (ret < 0)
 					ilog(LOG_ERR, "Failed to seek to beginning of media file");
-				ret = av_read_frame(mp->fmtctx, &mp->pkt);
+				ret = av_read_frame(mp->fmtctx, mp->pkt);
 			} else {
 				ilog(LOG_DEBUG, "EOF reading from media stream");
 				return;
@@ -427,26 +435,26 @@ static void media_player_read_packet(struct media_player *mp) {
 
 	// scale pts and duration according to sample rate
 
-	long long duration_scaled = mp->pkt.duration * avs->CODECPAR->sample_rate
+	long long duration_scaled = mp->pkt->duration * avs->CODECPAR->sample_rate
 		* avs->time_base.num / avs->time_base.den;
-	unsigned long long pts_scaled = mp->pkt.pts * avs->CODECPAR->sample_rate
+	unsigned long long pts_scaled = mp->pkt->pts * avs->CODECPAR->sample_rate
 		* avs->time_base.num / avs->time_base.den;
 
-	long long us_dur = mp->pkt.duration * 1000000LL * avs->time_base.num / avs->time_base.den;
+	long long us_dur = mp->pkt->duration * 1000000LL * avs->time_base.num / avs->time_base.den;
 	ilog(LOG_DEBUG, "read media packet: pts %llu duration %lli (scaled %llu/%lli, %lli us), "
 			"sample rate %i, time_base %i/%i",
-			(unsigned long long) mp->pkt.pts,
-			(long long) mp->pkt.duration,
+			(unsigned long long) mp->pkt->pts,
+			(long long) mp->pkt->duration,
 			pts_scaled,
 			duration_scaled,
 			us_dur,
 			avs->CODECPAR->sample_rate,
 			avs->time_base.num, avs->time_base.den);
 
-	media_player_add_packet(mp, (char *) mp->pkt.data, mp->pkt.size, us_dur, pts_scaled);
+	media_player_add_packet(mp, (char *) mp->pkt->data, mp->pkt->size, us_dur, pts_scaled);
 
 out:
-	av_packet_unref(&mp->pkt);
+	av_packet_unref(mp->pkt);
 }
 
 
