@@ -58,6 +58,7 @@ static void dump_packet_proc(struct media_packet *mp, const str *s);
 static void init_stream_proc(struct packet_stream *);
 static void setup_stream_proc(struct packet_stream *);
 static void setup_media_proc(struct call_media *);
+static void setup_monologue_proc(struct call_monologue *);
 static void kernel_info_proc(struct packet_stream *, struct rtpengine_target_info *);
 
 static void rec_pcap_eth_header(unsigned char *, struct packet_stream *);
@@ -92,6 +93,7 @@ static const struct recording_method methods[] = {
 		.init_stream_struct = init_stream_proc,
 		.setup_stream = setup_stream_proc,
 		.setup_media = setup_media_proc,
+		.setup_monologue = setup_monologue_proc,
 		.stream_kernel_info = kernel_info_proc,
 	},
 };
@@ -244,6 +246,12 @@ static void update_metadata(struct call *call, str *metadata) {
 	}
 }
 
+static void update_output_dest(struct call *call, str *output_dest) {
+	if (!output_dest || !output_dest->s || !call->recording)
+		return;
+	recording_meta_chunk(call->recording, "OUTPUT_DESTINATION", output_dest);
+}
+
 // lock must be held
 static void update_flags_proc(struct call *call) {
 	append_meta_chunk_null(call->recording, "RECORDING %u", call->recording_on ? 1 : 0);
@@ -259,9 +267,11 @@ static void recording_update_flags(struct call *call) {
 }
 
 // lock must be held
-void recording_start(struct call *call, const char *prefix, str *metadata) {
+void recording_start(struct call *call, const char *prefix, str *metadata, str *output_dest) {
 	update_metadata(call, metadata);
 
+	update_output_dest(call, output_dest);
+
 	if (call->recording) {
 		// already active
 		recording_update_flags(call);
@@ -294,6 +304,10 @@ void recording_start(struct call *call, const char *prefix, str *metadata) {
 	// function is called right at the start of the call, all of the following
 	// is essentially a no-op
 	GList *l;
+	for (l = call->monologues.head; l; l = l->next) {
+		struct call_monologue *ml = l->data;
+		recording_setup_monologue(ml);
+	}
 	for (l = call->medias.head; l; l = l->next) {
 		struct call_media *m = l->data;
 		recording_setup_media(m);
@@ -302,7 +316,7 @@ void recording_start(struct call *call, const char *prefix, str *metadata) {
 		struct packet_stream *ps = l->data;
 		recording_setup_stream(ps);
 		__unkernelize(ps);
-		ps->handler = NULL;
+		__reset_sink_handlers(ps);
 	}
 
 	recording_update_flags(call);
@@ -350,7 +364,7 @@ void detect_setup_recording(struct call *call, const str *recordcall, str *metad
 
 	if (!str_cmp(recordcall, "yes") || !str_cmp(recordcall, "on")) {
 		call->recording_on = 1;
-		recording_start(call, NULL, NULL);
+		recording_start(call, NULL, NULL, NULL);
 	}
 	else if (!str_cmp(recordcall, "no") || !str_cmp(recordcall, "off")) {
 		call->recording_on = 0;
@@ -425,7 +439,7 @@ static void sdp_after_pcap(struct recording *recording, GString *str, struct cal
 	// descriptor does not. Make sure to flush any unwritten contents
 	// so the file contents appear in order.
 	if (ml->label.len) {
-		fprintf(meta_fp, "\nLabel: %*s", ml->label.len, ml->label.s);
+		fprintf(meta_fp, "\nLabel: " STR_FORMAT, STR_FMT(&ml->label));
 	}
 	fprintf(meta_fp, "\nTimestamp started ms: ");
 	fprintf(meta_fp, "%.3lf", ml->started.tv_sec*1000.0+ml->started.tv_usec/1000.0);
@@ -456,16 +470,22 @@ static int rec_pcap_meta_finish_file(struct call *call) {
 	time_t start = call->created.tv_sec;
 	time_t end = rtpe_now.tv_sec;
 	char timebuffer[20];
-	struct tm *timeinfo;
+	struct tm timeinfo;
 	struct timeval *terminate;
 	terminate = &(((struct call_monologue *)call->monologues.head->data)->terminated);
 	fprintf(recording->u.pcap.meta_fp, "\nTimestamp terminated ms(first monologue): %.3lf", terminate->tv_sec*1000.0 + terminate->tv_usec/1000.0);
-	timeinfo = localtime(&start);
-	strftime(timebuffer, 20, "%FT%T", timeinfo);
-	fprintf(recording->u.pcap.meta_fp, "\n\ncall start time: %s\n", timebuffer);
-	timeinfo = localtime(&end);
-	strftime(timebuffer, 20, "%FT%T", timeinfo);
-	fprintf(recording->u.pcap.meta_fp, "call end time: %s\n", timebuffer);
+	if (localtime_r(&start, &timeinfo) == NULL) {
+		ilog(LOG_ERROR, "Cannot get start local time, while cleaning up recording meta file: %s", strerror(errno));
+	} else {
+		strftime(timebuffer, 20, "%FT%T", &timeinfo);
+		fprintf(recording->u.pcap.meta_fp, "\n\ncall start time: %s\n", timebuffer);
+	}
+	if (localtime_r(&end, &timeinfo) == NULL) {
+		ilog(LOG_ERROR, "Cannot get end local time, while cleaning up recording meta file: %s", strerror(errno));
+	} else {
+		strftime(timebuffer, 20, "%FT%T", &timeinfo);
+		fprintf(recording->u.pcap.meta_fp, "call end time: %s\n", timebuffer);
+	}
 
 	// Print metadata
 	if (call->metadata.len)
@@ -732,9 +752,6 @@ static void proc_init(struct call *call) {
 static void sdp_before_proc(struct recording *recording, const str *sdp, struct call_monologue *ml,
 		enum call_opmode opmode)
 {
-	append_meta_chunk_str(recording, &ml->tag, "TAG %u", ml->unique_id);
-	if (ml->label.len)
-		append_meta_chunk_str(recording, &ml->label, "LABEL %u", ml->unique_id);
 	append_meta_chunk_str(recording, sdp,
 			"SDP from %u before %s", ml->unique_id, get_opmode_text(opmode));
 }
@@ -798,6 +815,18 @@ static void setup_stream_proc(struct packet_stream *stream) {
 	append_meta_chunk(recording, buf, len, "STREAM %u interface", stream->unique_id);
 }
 
+static void setup_monologue_proc(struct call_monologue *ml) {
+	struct call *call = ml->call;
+	struct recording *recording = call->recording;
+
+	if (!recording)
+		return;
+
+	append_meta_chunk_str(recording, &ml->tag, "TAG %u", ml->unique_id);
+	if (ml->label.len)
+		append_meta_chunk_str(recording, &ml->label, "LABEL %u", ml->unique_id);
+}
+
 static void setup_media_proc(struct call_media *media) {
 	struct call *call = media->call;
 	struct recording *recording = call->recording;
@@ -807,7 +836,7 @@ static void setup_media_proc(struct call_media *media) {
 
 	append_meta_chunk_null(recording, "MEDIA %u PTIME %i", media->unique_id, media->ptime);
 
-	GList *pltypes = g_hash_table_get_values(media->codecs_recv);
+	GList *pltypes = g_hash_table_get_values(media->codecs.codecs);
 
 	for (GList *l = pltypes; l; l = l->next) {
 		struct rtp_payload_type *pt = l->data;
