@@ -22,7 +22,8 @@ int dtmf_event(struct media_packet *, str *, int);
 int dtmf_event_payload(str *, uint64_t *, uint64_t, struct dtmf_event *, GQueue *);
 void dtmf_event_free(void *);
 int dtmf_code_from_char(char);
-const char *dtmf_inject(struct call_media *media, int code, int volume, int duration, int pause);
+const char *dtmf_inject(struct call_media *media, int code, int volume, int duration, int pause,
+		struct call_media *sink);
 int dtmf_do_logging(void);
 
 
