@@ -57,11 +57,17 @@ static int ilbc_decoder_input(decoder_t *dec, const str *data, GQueue *out);
 static const char *dtmf_decoder_init(decoder_t *, const str *, const str *);
 static int dtmf_decoder_input(decoder_t *dec, const str *data, GQueue *out);
 
+static const char *cn_decoder_init(decoder_t *, const str *, const str *);
 static int cn_decoder_input(decoder_t *dec, const str *data, GQueue *out);
 
 static int format_cmp_ignore(const struct rtp_payload_type *, const struct rtp_payload_type *);
 
-static int amr_packet_lost(decoder_t *, GQueue *);
+static int generic_silence_dtx(decoder_t *, GQueue *, int);
+static int amr_dtx(decoder_t *, GQueue *, int);
+
+static int generic_cn_dtx_init(decoder_t *);
+static void generic_cn_dtx_cleanup(decoder_t *);
+static int generic_cn_dtx(decoder_t *, GQueue *, int);
 
 
 
@@ -100,11 +106,26 @@ static const codec_type_t codec_type_dtmf = {
 };
 static const codec_type_t codec_type_cn = {
 	.def_init = avc_def_init,
-	.decoder_init = avc_decoder_init,
+	.decoder_init = cn_decoder_init,
 	.decoder_input = cn_decoder_input,
 	.decoder_close = avc_decoder_close,
 };
 
+static const dtx_method_t dtx_method_silence = {
+	.method_id = DTX_SILENCE,
+	.do_dtx = generic_silence_dtx,
+};
+static const dtx_method_t dtx_method_cn = {
+	.method_id = DTX_CN,
+	.do_dtx = generic_cn_dtx,
+	.init = generic_cn_dtx_init,
+	.cleanup = generic_cn_dtx_cleanup,
+};
+static const dtx_method_t dtx_method_amr = {
+	.method_id = DTX_NATIVE,
+	.do_dtx = amr_dtx,
+};
+
 #ifdef HAVE_BCG729
 static packetizer_f packetizer_g729; // aggregate some frames into packets
 
@@ -141,6 +162,11 @@ static codec_def_t __codec_defs[] = {
 		.bits_per_sample = 8,
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
+		.silence_pattern = STR_CONST_INIT("\xd5"),
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 	{
 		.rtpname = "PCMU",
@@ -153,6 +179,11 @@ static codec_def_t __codec_defs[] = {
 		.bits_per_sample = 8,
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
+		.silence_pattern = STR_CONST_INIT("\xff"),
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 	{
 		.rtpname = "G723",
@@ -165,6 +196,10 @@ static codec_def_t __codec_defs[] = {
 		.packetizer = packetizer_passthrough,
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 	{
 		.rtpname = "G722",
@@ -177,6 +212,11 @@ static codec_def_t __codec_defs[] = {
 		.bits_per_sample = 8,
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
+		.silence_pattern = STR_CONST_INIT("\xfa"),
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 	{
 		.rtpname = "QCELP",
@@ -186,6 +226,10 @@ static codec_def_t __codec_defs[] = {
 		.packetizer = packetizer_passthrough,
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 #ifndef HAVE_BCG729
 	{
@@ -198,6 +242,10 @@ static codec_def_t __codec_defs[] = {
 		.packetizer = packetizer_passthrough,
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 	{
 		.rtpname = "G729a",
@@ -209,6 +257,10 @@ static codec_def_t __codec_defs[] = {
 		.packetizer = packetizer_passthrough,
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 #else
 	{
@@ -223,6 +275,10 @@ static codec_def_t __codec_defs[] = {
 		.bits_per_sample = 1, // 10 ms frame has 80 samples and encodes as (max) 10 bytes = 80 bits
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_bcg729,
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 	{
 		.rtpname = "G729a",
@@ -235,6 +291,10 @@ static codec_def_t __codec_defs[] = {
 		.bits_per_sample = 1, // 10 ms frame has 80 samples and encodes as (max) 10 bytes = 80 bits
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_bcg729,
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 #endif
 	{
@@ -247,6 +307,10 @@ static codec_def_t __codec_defs[] = {
 		.packetizer = packetizer_passthrough,
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 	{
 		.rtpname = "GSM",
@@ -258,6 +322,10 @@ static codec_def_t __codec_defs[] = {
 		.packetizer = packetizer_passthrough,
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 	{
 		.rtpname = "iLBC",
@@ -286,7 +354,12 @@ static codec_def_t __codec_defs[] = {
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
 		.init = opus_init,
+		.format_cmp = format_cmp_ignore,
 		.set_enc_options = opus_set_enc_options,
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 	{
 		.rtpname = "vorbis",
@@ -297,6 +370,10 @@ static codec_def_t __codec_defs[] = {
 		.packetizer = packetizer_passthrough,
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 	{
 		.rtpname = "ac3",
@@ -305,6 +382,10 @@ static codec_def_t __codec_defs[] = {
 		.packetizer = packetizer_passthrough,
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 	{
 		.rtpname = "eac3",
@@ -313,6 +394,10 @@ static codec_def_t __codec_defs[] = {
 		.packetizer = packetizer_passthrough,
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 	{
 		.rtpname = "ATRAC3",
@@ -321,6 +406,10 @@ static codec_def_t __codec_defs[] = {
 		.packetizer = packetizer_passthrough,
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 	{
 		.rtpname = "ATRAC-X",
@@ -329,6 +418,10 @@ static codec_def_t __codec_defs[] = {
 		.packetizer = packetizer_passthrough,
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 #if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(57, 0, 0)
 	{
@@ -338,6 +431,10 @@ static codec_def_t __codec_defs[] = {
 		.packetizer = packetizer_passthrough,
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 	{
 		.rtpname = "EVRC0",
@@ -347,6 +444,10 @@ static codec_def_t __codec_defs[] = {
 		.packetizer = packetizer_passthrough,
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 	{
 		.rtpname = "EVRC1",
@@ -356,6 +457,10 @@ static codec_def_t __codec_defs[] = {
 		.packetizer = packetizer_passthrough,
 		.media_type = MT_AUDIO,
 		.codec_type = &codec_type_avcodec,
+		.dtx_methods = {
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 #endif
 	{
@@ -374,7 +479,12 @@ static codec_def_t __codec_defs[] = {
 		.codec_type = &codec_type_amr,
 		.set_enc_options = amr_set_enc_options,
 		.set_dec_options = amr_set_dec_options,
-		.packet_lost = amr_packet_lost,
+		.amr = 1,
+		.dtx_methods = {
+			[DTX_NATIVE] = &dtx_method_amr,
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 	{
 		.rtpname = "AMR-WB",
@@ -392,7 +502,12 @@ static codec_def_t __codec_defs[] = {
 		.codec_type = &codec_type_amr,
 		.set_enc_options = amr_set_enc_options,
 		.set_dec_options = amr_set_dec_options,
-		.packet_lost = amr_packet_lost,
+		.amr = 1,
+		.dtx_methods = {
+			[DTX_NATIVE] = &dtx_method_amr,
+			[DTX_SILENCE] = &dtx_method_silence,
+			[DTX_CN] = &dtx_method_cn,
+		},
 	},
 	{
 		.rtpname = "telephone-event",
@@ -449,6 +564,7 @@ static codec_def_t __codec_defs[] = {
 
 static GQueue __supplemental_codecs = G_QUEUE_INIT;
 const GQueue * const codec_supplemental_codecs = &__supplemental_codecs;
+static codec_def_t *codec_def_cn;
 
 
 
@@ -478,6 +594,8 @@ static const char *avc_decoder_init(decoder_t *dec, const str *fmtp, const str *
 	if (!codec)
 		return "codec not supported";
 
+	dec->u.avc.avpkt = av_packet_alloc();
+
 	dec->u.avc.avcctx = avcodec_alloc_context3(codec);
 	if (!dec->u.avc.avcctx)
 		return "failed to alloc codec context";
@@ -529,23 +647,33 @@ decoder_t *decoder_new_fmtp(const codec_def_t *def, int clockrate, int channels,
 	ret->in_format.channels = channels;
 	ret->in_format.clockrate = clockrate;
 	// output defaults to same as input
-	ret->out_format = ret->in_format;
+	ret->dest_format = ret->in_format;
+	ret->dec_out_format = ret->in_format;
 	if (resample_fmt)
-		ret->out_format = *resample_fmt;
+		ret->dest_format = *resample_fmt;
 	if (ptime > 0)
 		ret->ptime = ptime;
 	else
 		ret->ptime = def->default_ptime;
 
+	// init with first supported DTX method
+	enum dtx_method dm = -1;
+	for (int i = 0; i < NUM_DTX_METHODS; i++) {
+		if (def->dtx_methods[i]) {
+			dm = i;
+			break;
+		}
+	}
+
 	err = def->codec_type->decoder_init(ret, fmtp, extra_opts);
 	if (err)
 		goto err;
 
-	av_init_packet(&ret->u.avc.avpkt);
-
 	ret->pts = (uint64_t) -1LL;
 	ret->rtp_ts = (unsigned long) -1L;
 
+	decoder_switch_dtx(ret, dm);
+
 	return ret;
 
 err:
@@ -557,6 +685,40 @@ err:
 }
 
 
+int decoder_switch_dtx(decoder_t *dec, enum dtx_method dm) {
+	if (dec->dtx.cleanup)
+		dec->dtx.cleanup(dec);
+	ZERO(dec->dtx);
+	unsigned int i = dm;
+	if (i >= NUM_DTX_METHODS)
+		return -1;
+	const dtx_method_t *dmp = dec->def->dtx_methods[i];
+	if (!dmp)
+		return -1;
+	dec->dtx = *dmp;
+	if (dmp->init) {
+		if (dmp->init(dec)) {
+			ilog(LOG_ERR, "Failed to initialise DTX (%u)", i);
+			decoder_switch_dtx(dec, -1);
+			return -1;
+		}
+	}
+	return 0;
+}
+
+int decoder_set_cn_dtx(decoder_t *dec, const str *cn_pl) {
+	if (decoder_switch_dtx(dec, DTX_CN))
+		return -1;
+	dec->dtx.u.cn.cn_payload = cn_pl;
+	return 0;
+}
+
+
+gboolean decoder_has_dtx(decoder_t *dec) {
+	return dec->dtx.do_dtx == NULL ? FALSE : TRUE;
+}
+
+
 static void avc_decoder_close(decoder_t *dec) {
 #if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(56, 1, 0)
 	avcodec_free_context(&dec->u.avc.avcctx);
@@ -564,6 +726,7 @@ static void avc_decoder_close(decoder_t *dec) {
 	avcodec_close(dec->u.avc.avcctx);
 	av_free(dec->u.avc.avcctx);
 #endif
+	av_packet_free(&dec->u.avc.avpkt);
 }
 
 
@@ -575,6 +738,8 @@ void decoder_close(decoder_t *dec) {
 	if (dec->def && dec->def->codec_type && dec->def->codec_type->decoder_close)
 		dec->def->codec_type->decoder_close(dec);
 
+	decoder_switch_dtx(dec, -1);
+
 	resample_shutdown(&dec->resampler);
 	g_slice_free1(sizeof(*dec), dec);
 }
@@ -584,9 +749,9 @@ static int avc_decoder_input(decoder_t *dec, const str *data, GQueue *out) {
 	const char *err;
 	int av_ret = 0;
 
-	dec->u.avc.avpkt.data = (unsigned char *) data->s;
-	dec->u.avc.avpkt.size = data->len;
-	dec->u.avc.avpkt.pts = dec->pts;
+	dec->u.avc.avpkt->data = (unsigned char *) data->s;
+	dec->u.avc.avpkt->size = data->len;
+	dec->u.avc.avpkt->pts = dec->pts;
 
 	AVFrame *frame = NULL;
 
@@ -601,13 +766,13 @@ static int avc_decoder_input(decoder_t *dec, const str *data, GQueue *out) {
 			goto err;
 
 #if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(57, 36, 0)
-		if (dec->u.avc.avpkt.size) {
-			av_ret = avcodec_send_packet(dec->u.avc.avcctx, &dec->u.avc.avpkt);
+		if (dec->u.avc.avpkt->size) {
+			av_ret = avcodec_send_packet(dec->u.avc.avcctx, dec->u.avc.avpkt);
 			cdbg("send packet ret %i", av_ret);
 			err = "failed to send packet to avcodec";
 			if (av_ret == 0) {
 				// consumed the packet
-				dec->u.avc.avpkt.size = 0;
+				dec->u.avc.avpkt->size = 0;
 				keep_going = 1;
 			}
 			else {
@@ -634,10 +799,10 @@ static int avc_decoder_input(decoder_t *dec, const str *data, GQueue *out) {
 		}
 #else
 		// only do this if we have any input left
-		if (dec->u.avc.avpkt.size == 0)
+		if (dec->u.avc.avpkt->size == 0)
 			break;
 
-		av_ret = avcodec_decode_audio4(dec->u.avc.avcctx, frame, &got_frame, &dec->u.avc.avpkt);
+		av_ret = avcodec_decode_audio4(dec->u.avc.avcctx, frame, &got_frame, dec->u.avc.avpkt);
 		cdbg("decode frame ret %i, got frame %i", av_ret, got_frame);
 		err = "failed to decode audio packet";
 		if (av_ret < 0)
@@ -645,10 +810,10 @@ static int avc_decoder_input(decoder_t *dec, const str *data, GQueue *out) {
 		if (av_ret > 0) {
 			// consumed some input
 			err = "invalid return value";
-			if (av_ret > dec->u.avc.avpkt.size)
+			if (av_ret > dec->u.avc.avpkt->size)
 				goto err;
-			dec->u.avc.avpkt.size -= av_ret;
-			dec->u.avc.avpkt.data += av_ret;
+			dec->u.avc.avpkt->size -= av_ret;
+			dec->u.avc.avpkt->data += av_ret;
 			keep_going = 1;
 		}
 		if (got_frame)
@@ -663,8 +828,8 @@ static int avc_decoder_input(decoder_t *dec, const str *data, GQueue *out) {
 			frame->pts = frame->pkt_pts;
 #endif
 			if (G_UNLIKELY(frame->pts == AV_NOPTS_VALUE))
-				frame->pts = dec->u.avc.avpkt.pts;
-			dec->u.avc.avpkt.pts += frame->nb_samples;
+				frame->pts = dec->u.avc.avpkt->pts;
+			dec->u.avc.avpkt->pts += frame->nb_samples;
 
 			g_queue_push_tail(out, frame);
 			frame = NULL;
@@ -682,7 +847,7 @@ err:
 	return -1;
 }
 
-static int __decoder_input_data(decoder_t *dec, const str *data, unsigned long ts,
+static int __decoder_input_data(decoder_t *dec, const str *data, unsigned long ts, int *ptime,
 		int (*callback)(decoder_t *, AVFrame *, void *u1, void *u2), void *u1, void *u2)
 {
 	GQueue frames = G_QUEUE_INIT;
@@ -690,7 +855,7 @@ static int __decoder_input_data(decoder_t *dec, const str *data, unsigned long t
 	if (G_UNLIKELY(!dec))
 		return -1;
 
-	if (!data && !dec->def->packet_lost)
+	if (!data && (!dec->dtx.do_dtx || !ptime))
 		return 0;
 
 	ts *= dec->def->clockrate_mult;
@@ -704,7 +869,7 @@ static int __decoder_input_data(decoder_t *dec, const str *data, unsigned long t
 	}
 	else {
 		// shift pts according to rtp ts shift
-		u_int64_t shift_ts = ts - dec->rtp_ts;
+		uint64_t shift_ts = ts - dec->rtp_ts;
 		if ((shift_ts * 1000) / dec->in_format.clockrate > PACKET_TS_RESET_THRES) {
 			ilog(LOG_DEBUG, "Timestamp discontinuity detected, resetting timestamp from "
 					"%lu to %lu",
@@ -719,12 +884,15 @@ static int __decoder_input_data(decoder_t *dec, const str *data, unsigned long t
 	if (data)
 		dec->def->codec_type->decoder_input(dec, data, &frames);
 	else
-		dec->def->packet_lost(dec, &frames);
+		dec->dtx.do_dtx(dec, &frames, *ptime);
 
 	AVFrame *frame;
 	int ret = 0;
+	unsigned long samples = 0;
 	while ((frame = g_queue_pop_head(&frames))) {
-		AVFrame *rsmp_frame = resample_frame(&dec->resampler, frame, &dec->out_format);
+		samples += frame->nb_samples;
+		dec->dec_out_format.format = frame->format;
+		AVFrame *rsmp_frame = resample_frame(&dec->resampler, frame, &dec->dest_format);
 		if (!rsmp_frame) {
 			ilog(LOG_ERR | LOG_FLAG_LIMIT, "Resampling failed");
 			ret = -1;
@@ -736,6 +904,9 @@ static int __decoder_input_data(decoder_t *dec, const str *data, unsigned long t
 		av_frame_free(&frame);
 	}
 
+	if (ptime)
+		*ptime = samples * 1000L / dec->in_format.clockrate;
+
 	return ret;
 }
 int decoder_input_data(decoder_t *dec, const str *data, unsigned long ts,
@@ -743,12 +914,19 @@ int decoder_input_data(decoder_t *dec, const str *data, unsigned long ts,
 {
 	if (!data || !data->s || !data->len)
 		return 0;
-	return __decoder_input_data(dec, data, ts, callback, u1, u2);
+	return __decoder_input_data(dec, data, ts, NULL, callback, u1, u2);
 }
-int decoder_lost_packet(decoder_t *dec, unsigned long ts,
+int decoder_input_data_ptime(decoder_t *dec, const str *data, unsigned long ts, int *ptime,
 		int (*callback)(decoder_t *, AVFrame *, void *u1, void *u2), void *u1, void *u2)
 {
-	return __decoder_input_data(dec, NULL, ts, callback, u1, u2);
+	if (!data || !data->s || !data->len)
+		return 0;
+	return __decoder_input_data(dec, data, ts, ptime, callback, u1, u2);
+}
+int decoder_dtx(decoder_t *dec, unsigned long ts, int ptime,
+		int (*callback)(decoder_t *, AVFrame *, void *u1, void *u2), void *u1, void *u2)
+{
+	return __decoder_input_data(dec, NULL, ts, &ptime, callback, u1, u2);
 }
 
 
@@ -857,6 +1035,9 @@ void codeclib_init(int print) {
 		if (def->codec_type && def->codec_type->def_init)
 			def->codec_type->def_init(def);
 
+		if (!strcmp(def->rtpname, "CN"))
+			codec_def_cn = def;
+
 		if (print) {
 			if (def->support_encoding && def->support_decoding) {
 				if (def->default_channels > 0 && def->default_clockrate >= 0)
@@ -984,7 +1165,7 @@ static void *__packet_sequencer_next_packet(packet_sequencer_t *ps, int num_wait
 
 out:
 	;
-	u_int16_t l = packet->seq - ps->seq;
+	uint16_t l = packet->seq - ps->seq;
 	ps->lost_count += l;
 
 	g_tree_steal(ps->packets, GINT_TO_POINTER(packet->seq));
@@ -1054,6 +1235,7 @@ encoder_t *encoder_new(void) {
 	encoder_t *ret = g_slice_alloc0(sizeof(*ret));
 	format_init(&ret->requested_format);
 	format_init(&ret->actual_format);
+	ret->avpkt = av_packet_alloc();
 	return ret;
 }
 
@@ -1126,6 +1308,9 @@ int encoder_config_fmtp(encoder_t *enc, const codec_def_t *def, int bitrate, int
 
 	encoder_close(enc);
 
+	if (ptime <= 0)
+		ptime = 20;
+
 	enc->requested_format = *requested_format;
 	enc->def = def;
 	enc->ptime = ptime / def->clockrate_mult;
@@ -1135,8 +1320,6 @@ int encoder_config_fmtp(encoder_t *enc, const codec_def_t *def, int bitrate, int
 	if (err)
 		goto err;
 
-	av_init_packet(&enc->avpkt);
-
 // output frame and fifo
 	enc->frame = av_frame_alloc();
 
@@ -1194,6 +1377,7 @@ void encoder_close(encoder_t *enc) {
 }
 void encoder_free(encoder_t *enc) {
 	encoder_close(enc);
+	av_packet_free(&enc->avpkt);
 	g_slice_free1(sizeof(*enc), enc);
 }
 
@@ -1222,7 +1406,7 @@ static int avc_encoder_input(encoder_t *enc, AVFrame **frame) {
 		}
 	}
 
-	av_ret = avcodec_receive_packet(enc->u.avc.avcctx, &enc->avpkt);
+	av_ret = avcodec_receive_packet(enc->u.avc.avcctx, enc->avpkt);
 	cdbg("receive packet ret %i", av_ret);
 	if (av_ret == 0) {
 		// got some data
@@ -1239,7 +1423,7 @@ static int avc_encoder_input(encoder_t *enc, AVFrame **frame) {
 	if (!*frame)
 		return 0;
 
-	av_ret = avcodec_encode_audio2(enc->u.avc.avcctx, &enc->avpkt, *frame, &got_packet);
+	av_ret = avcodec_encode_audio2(enc->u.avc.avcctx, enc->avpkt, *frame, &got_packet);
 	cdbg("encode frame ret %i, got packet %i", av_ret, got_packet);
 	if (av_ret == 0)
 		*frame = NULL; // consumed
@@ -1252,16 +1436,16 @@ static int avc_encoder_input(encoder_t *enc, AVFrame **frame) {
 	if (!got_packet)
 		return keep_going;
 
-	cdbg("output avpkt size is %i", (int) enc->avpkt.size);
-	cdbg("output pkt pts/dts is %li/%li", (long) enc->avpkt.pts,
-			(long) enc->avpkt.dts);
+	cdbg("output avpkt size is %i", (int) enc->avpkt->size);
+	cdbg("output pkt pts/dts is %li/%li", (long) enc->avpkt->pts,
+			(long) enc->avpkt->dts);
 
 	// the encoder may return frames with the same dts multiple consecutive times.
 	// the muxer may not like this, so ensure monotonically increasing dts.
-	if (enc->mux_dts > enc->avpkt.dts)
-		enc->avpkt.dts = enc->mux_dts;
-	if (enc->avpkt.pts < enc->avpkt.dts)
-		enc->avpkt.pts = enc->avpkt.dts;
+	if (enc->mux_dts > enc->avpkt->dts)
+		enc->avpkt->dts = enc->mux_dts;
+	if (enc->avpkt->pts < enc->avpkt->dts)
+		enc->avpkt->pts = enc->avpkt->dts;
 
 	return keep_going;
 
@@ -1274,9 +1458,11 @@ err:
 int encoder_input_data(encoder_t *enc, AVFrame *frame,
 		int (*callback)(encoder_t *, void *u1, void *u2), void *u1, void *u2)
 {
-	enc->avpkt.size = 0;
+	enc->avpkt->size = 0;
 
 	while (1) {
+		if (!enc->def || !enc->def->codec_type)
+			break;
 		if (!enc->def->codec_type->encoder_input)
 			break;
 
@@ -1284,22 +1470,22 @@ int encoder_input_data(encoder_t *enc, AVFrame *frame,
 		if (ret < 0)
 			return -1;
 
-		if (enc->avpkt.size) {
+		if (enc->avpkt->size) {
 			// don't rely on the encoder producing steady timestamps,
 			// instead keep track of them ourselves based on the returned
 			// frame duration
-			enc->avpkt.pts = enc->next_pts;
+			enc->avpkt->pts = enc->next_pts;
 
 			if (enc->def->codec_type->encoder_got_packet)
 				enc->def->codec_type->encoder_got_packet(enc);
 
 			callback(enc, u1, u2);
 
-			enc->next_pts += enc->avpkt.duration;
-			enc->mux_dts = enc->avpkt.dts + 1; // min next expected dts
+			enc->next_pts += enc->avpkt->duration;
+			enc->mux_dts = enc->avpkt->dts + 1; // min next expected dts
 
-			av_packet_unref(&enc->avpkt);
-			enc->avpkt.size = 0;
+			av_packet_unref(enc->avpkt);
+			enc->avpkt->size = 0;
 		}
 
 		if (ret == 0)
@@ -1374,7 +1560,7 @@ static int packetizer_samplestream(AVPacket *pkt, GString *buf, str *input_outpu
 	memcpy(input_output->s, buf->str, input_output->len);
 	g_string_erase(buf, 0, input_output->len);
 	// adjust output pts
-	enc->avpkt.pts = enc->packet_pts;
+	enc->avpkt->pts = enc->packet_pts;
 	enc->packet_pts += input_output->len * (enc->def->bits_per_sample * enc->def->clockrate_mult / 8);
 	return buf->len >= input_output->len ? 1 : 0;
 }
@@ -1456,7 +1642,7 @@ static void opus_init(struct rtp_payload_type *pt) {
 }
 
 static void opus_set_enc_options(encoder_t *enc, const str *fmtp, const str *codec_opts) {
-	if (enc->ptime)
+	if (enc->ptime > 0)
 		codeclib_set_av_opt_int(enc, "frame_duration", enc->ptime);
 	// XXX additional opus options
 }
@@ -1590,7 +1776,7 @@ static void codeclib_key_value_parse(const str *instr, int need_value,
 
 
 
-const static unsigned int amr_bitrates[AMR_FT_TYPES] = {
+static const unsigned int amr_bitrates[AMR_FT_TYPES] = {
 	4750, // 0
 	5150, // 1
 	5900, // 2
@@ -1606,7 +1792,7 @@ const static unsigned int amr_bitrates[AMR_FT_TYPES] = {
 	0, // invalid // 12
 	0, // invalid // 13
 };
-const static unsigned int amr_bits_per_frame[AMR_FT_TYPES] = {
+static const unsigned int amr_bits_per_frame[AMR_FT_TYPES] = {
 	95, // 4.75 kbit/s // 0
 	103, // 5.15 kbit/s // 1
 	118, // 5.90 kbit/s // 2
@@ -1622,7 +1808,7 @@ const static unsigned int amr_bits_per_frame[AMR_FT_TYPES] = {
 	0, // invalid // 12
 	0, // invalid // 13
 };
-const static unsigned int amr_wb_bitrates[AMR_FT_TYPES] = {
+static const unsigned int amr_wb_bitrates[AMR_FT_TYPES] = {
 	6600, // 0
 	8850, // 1
 	12650, // 2
@@ -1638,7 +1824,7 @@ const static unsigned int amr_wb_bitrates[AMR_FT_TYPES] = {
 	0, // invalid // 12
 	0, // invalid // 13
 };
-const static unsigned int amr_wb_bits_per_frame[AMR_FT_TYPES] = {
+static const unsigned int amr_wb_bits_per_frame[AMR_FT_TYPES] = {
 	132, // 6.60 kbit/s // 0
 	177, // 8.85 kbit/s // 1
 	253, // 12.65 kbit/s // 2
@@ -1844,7 +2030,7 @@ static int amr_decoder_input(decoder_t *dec, const str *data, GQueue *out) {
 		if (!dec->u.avc.u.amr.last_cmr.tv_sec) // start tracking now
 			dec->u.avc.u.amr.last_cmr = rtpe_now;
 		else if (timeval_diff(&rtpe_now, &dec->u.avc.u.amr.last_cmr)
-				>= dec->codec_options.amr.mode_change_interval * 1000) {
+				>= (long long) dec->codec_options.amr.mode_change_interval * 1000) {
 			// switch up if we can
 			decoder_event(dec, CE_AMR_CMR_RECV, GUINT_TO_POINTER(0xffff));
 			dec->u.avc.u.amr.last_cmr = rtpe_now;
@@ -1928,7 +2114,7 @@ static int amr_decoder_input(decoder_t *dec, const str *data, GQueue *out) {
 			goto err;
 
 		// add TOC byte
-		str_shift(&frame, -1);
+		str_unshift(&frame, 1);
 		frame.s[0] = toc_byte & 0x7c; // strip F bit, keep FT and Q, zero padding (01111100)
 
 		if (dec->codec_options.amr.octet_aligned && (bits % 8) != 0) {
@@ -1938,8 +2124,22 @@ static int amr_decoder_input(decoder_t *dec, const str *data, GQueue *out) {
 		}
 
 		err = "failed to decode AMR data";
-		if (avc_decoder_input(dec, &frame, out))
-			goto err;
+		if (bits == 40) {
+			// SID
+			if (dec->dtx.method_id == DTX_NATIVE) {
+				if (avc_decoder_input(dec, &frame, out))
+					goto err;
+			}
+			else {
+				// use the DTX generator to replace SID
+				if (dec->dtx.do_dtx(dec, out, 20))
+					goto err;
+			}
+		}
+		else {
+			if (avc_decoder_input(dec, &frame, out))
+				goto err;
+		}
 
 		amr_bitrate_tracker(dec, ft);
 	}
@@ -2113,7 +2313,8 @@ static int packetizer_amr(AVPacket *pkt, GString *buf, str *output, encoder_t *e
 
 	return 0;
 }
-static int amr_packet_lost(decoder_t *dec, GQueue *out) {
+static int amr_dtx(decoder_t *dec, GQueue *out, int ptime) {
+	// ignore ptime, must be 20
 	ilog(LOG_DEBUG, "pushing empty/lost frame to AMR decoder");
 	unsigned char frame_buf[1];
 	frame_buf[0] = 0xf << 3; // no data
@@ -2125,6 +2326,61 @@ static int amr_packet_lost(decoder_t *dec, GQueue *out) {
 
 
 
+static int generic_silence_dtx(decoder_t *dec, GQueue *out, int ptime) {
+	if (dec->dec_out_format.format == -1)
+		return -1;
+
+	if (ptime <= 0)
+		ptime = 20;
+	int num_samples = ptime * dec->in_format.clockrate / 1000;
+	ilog(LOG_DEBUG, "pushing %i silence samples into %s decoder", num_samples, dec->def->rtpname);
+
+	// create dummy frame, fill with silence, pretend it was returned from the decoder
+	AVFrame *frame = av_frame_alloc();
+	frame->nb_samples = num_samples;
+	frame->format = dec->dec_out_format.format;
+	frame->sample_rate = dec->dec_out_format.clockrate;
+	frame->channel_layout = av_get_default_channel_layout(dec->dec_out_format.channels);
+	if (av_frame_get_buffer(frame, 0) < 0) {
+		av_frame_free(&frame);
+		return -1;
+	}
+
+	memset(frame->extended_data[0], 0, frame->linesize[0]);
+
+	// advance PTS
+	frame->pts = dec->u.avc.avpkt->pts;
+	dec->u.avc.avpkt->pts += frame->nb_samples;
+
+	g_queue_push_tail(out, frame);
+
+	return 0;
+}
+
+
+static int cn_append_frame(decoder_t *dec, AVFrame *f, void *u1, void *u2) {
+	GQueue *out = u1;
+	g_queue_push_tail(out, f);
+	return 0;
+}
+
+static int generic_cn_dtx(decoder_t *dec, GQueue *out, int ptime) {
+	dec->dtx.u.cn.cn_dec->ptime = ptime;
+	return decoder_input_data(dec->dtx.u.cn.cn_dec, dec->dtx.u.cn.cn_payload,
+			dec->rtp_ts, cn_append_frame, out, NULL);
+}
+
+static int generic_cn_dtx_init(decoder_t *dec) {
+	dec->dtx.u.cn.cn_dec = decoder_new_fmt(codec_def_cn, 8000, 1, dec->ptime, &dec->dest_format);
+	return 0;
+}
+
+static void generic_cn_dtx_cleanup(decoder_t *dec) {
+	decoder_close(dec->dtx.u.cn.cn_dec);
+}
+
+
+
 
 #ifdef HAVE_BCG729
 static void bcg729_def_init(codec_def_t *def) {
@@ -2150,7 +2406,7 @@ static const char *bcg729_decoder_init(decoder_t *dec, const str *fmtp, const st
 
 static int bcg729_decoder_input(decoder_t *dec, const str *data, GQueue *out) {
 	str input = *data;
-	u_int64_t pts = dec->pts;
+	uint64_t pts = dec->pts;
 
 	while (input.len >= 2) {
 		int frame_len = input.len >= 10 ? 10 : 2;
@@ -2208,17 +2464,18 @@ static int bcg729_encoder_input(encoder_t *enc, AVFrame **frame) {
 		return -1;
 	}
 
-	av_new_packet(&enc->avpkt, 10);
+	av_new_packet(enc->avpkt, 10);
 	unsigned char len = 0;
 
-	bcg729Encoder(enc->u.bcg729, (void *) (*frame)->extended_data[0], enc->avpkt.data, &len);
+	bcg729Encoder(enc->u.bcg729, (void *) (*frame)->extended_data[0], enc->avpkt->data, &len);
 	if (!len) {
-		av_packet_unref(&enc->avpkt);
+		av_packet_unref(enc->avpkt);
 		return 0;
 	}
 
-	enc->avpkt.size = len;
-	enc->avpkt.pts = (*frame)->pts;
+	enc->avpkt->size = len;
+	enc->avpkt->pts = (*frame)->pts;
+	enc->avpkt->duration = len * 8; // Duration is used by encoder_input_data for pts calculation
 
 	return 0;
 }
@@ -2284,7 +2541,7 @@ static const char *dtmf_decoder_init(decoder_t *dec, const str *fmtp, const str
 static int dtmf_decoder_input(decoder_t *dec, const str *data, GQueue *out) {
 	struct telephone_event_payload *dtmf;
 	if (data->len < sizeof(*dtmf)) {
-		ilog(LOG_WARN | LOG_FLAG_LIMIT, "Short DTMF event packet (len %u)", data->len);
+		ilog(LOG_WARN | LOG_FLAG_LIMIT, "Short DTMF event packet (len %zu)", data->len);
 		return -1;
 	}
 	dtmf = (void *) data->s;
@@ -2345,10 +2602,17 @@ static int format_cmp_ignore(const struct rtp_payload_type *a, const struct rtp_
 
 
 
+static const char *cn_decoder_init(decoder_t *dec, const str *fmtp, const str *opts) {
+	// the ffmpeg cngdec always runs at 8000
+	dec->in_format.clockrate = 8000;
+	dec->in_format.channels = 1;
+	dec->resampler.no_filter = true;
+	return avc_decoder_init(dec, fmtp, opts);
+}
 static int cn_decoder_input(decoder_t *dec, const str *data, GQueue *out) {
 	// generate one set of ptime worth of samples
 	int ptime = dec->ptime;
-	if (!ptime)
+	if (ptime <= 0)
 		ptime = 20; // ?
 	int samples = dec->in_format.clockrate * ptime / 1000;
 	dec->u.avc.avcctx->frame_size = samples;
